
#include "mutex.h"

BEGIN_HSPS_NAMESPACE

void StaticMutex::compute()
{
  pairix.init(instance.n_atoms());
  tab.assign_value(false, pairix.n_pairs());
  for (index_type i = 0; i < instance.init_atoms.size(); i++)
    for (index_type j = i; j < instance.init_atoms.size(); j++)
      tab[pairix(instance.init_atoms[i], instance.init_atoms[j])] = true;

  bool done = false;
  while (!done) {
#ifdef TRACE_PRINT_LOTS
    std::cerr << "begin iteration" << std::endl;
#endif
    done = true;
    for (index_type a = 0; a < instance.n_actions(); a++) {
      Instance::Action& act = instance.actions[a];
      if (eval(act.pre)) {
#ifdef TRACE_PRINT_LOTS
	std::cerr << "applying action " << a << "."
		  << instance.actions[a].name << std::endl;
#endif
	for (index_type i = 0; i < instance.n_atoms(); i++) {
	  // if i in act.add...
	  if (act.add.contains(i)) {
	    // update i
	    if (!tab[pairix(i, i)]) {
#ifdef TRACE_PRINT_LOTS
	      std::cerr << "updating " << i << "."
			<< instance.atoms[i].name << std::endl;
#endif
	      tab[pairix(i, i)] = true;
	      done = false;
	    }
	    for (index_type j = 0; j < act.add.length(); j++)
	      // update {i,q} for every q in act.add
	      if (act.add[j] > i) // q < i have already been updated (as i)
		if (!tab[pairix(act.add[j], i)]) {
#ifdef TRACE_PRINT_LOTS
		  std::cerr << "updating " << i << "."
			    << instance.atoms[i].name
			    << ", " << act.add[j]
			    << instance.atoms[act.add[j]].name
			    << std::endl;
#endif
		  tab[pairix(act.add[j], i)] = true;
		  done = false;
		}
	  }
	  // else if i not in act.del (i.e. case act + noop(i))...
	  else if (tab[pairix(i, i)] && !act.del.contains(i)) {
	    // compute H(act.pre U {i})
	    bool ok = true;
	    for (index_type k = 0; (k < act.pre.length()) && ok; k++)
	      if (!tab[pairix(i, act.pre[k])])
		ok = false;
	    // update {i,q} for every q in act.add
	    if (ok) {
	      for (index_type j = 0; j < act.add.length(); j++)
		if (!tab[pairix(i, act.add[j])]) {
#ifdef TRACE_PRINT_LOTS
		  std::cerr << "updating " << i << "."
			    << instance.atoms[i].name
			    << ", " << act.add[j]
			    << instance.atoms[act.add[j]].name
			    << std::endl;
#endif
		  tab[pairix(i, act.add[j])] = true;
		  done = false;
		}
	    }
	  }
	} // for each atom i
      } // if (eval(act.pre))
    } // for each action
#ifdef TRACE_PRINT_LOTS
    std::cerr << "end iteration, done = " << done << std::endl;
#endif
  }
}

StaticMutex::StaticMutex(Instance& ins) : instance(ins)
{
  compute();
}

bool StaticMutex::eval(index_type i, index_type j) const
{
  return tab[pairix(i, j)];
}

bool StaticMutex::eval(const index_set& s) const
{
#ifdef TRACE_PRINT_LOTS
  std::cerr << "eval " << s << std::endl;
#endif
  for (index_type i = 0; i < s.size(); i++)
    for (index_type j = i; j < s.size(); j++)
      if (!tab[pairix(s[i], s[j])]) {
#ifdef TRACE_PRINT_LOTS
	std::cerr << " - (" << s[i] << "," << s[j] << ") = pair #"
		  << pairix(s[i], s[j]) << " is false" << std::endl;
#endif
	return false;
      }
#ifdef TRACE_PRINT_LOTS
  std::cerr << " - return true" << std::endl;
#endif
  return true;
}

bool StaticMutex::eval(const index_set& s, index_type i) const
{
  if (!tab[pairix(i, i)]) return false;
  if (!eval(s)) return false;
  for (index_type k = 0; k < s.size(); k++)
    if (!tab[pairix(s[k], i)])
      return false;
  return true;
}

bool StaticMutex::eval(const index_set& s1, const index_set& s2) const
{
  if (!eval(s1)) return false;
  if (!eval(s2)) return false;
  for (index_type k1 = 0; k1 < s1.size(); k1++)
    for (index_type k2 = 0; k2 < s2.size(); k2++)
      if (!tab[pairix(s1[k1], s2[k2])])
	return false;
  return true;
}

END_HSPS_NAMESPACE
